import Flutter
import UIKit
import kameleoonClient

public class SwiftKameleoonClientFlutterPlugin: NSObject, FlutterPlugin {
    private var clients = [String : KameleoonClient]()
    private var channel : FlutterMethodChannel

    private init(channel: FlutterMethodChannel) {
        self.channel = channel
    }

    public static func register(with registrar: FlutterPluginRegistrar) {
        let channel = FlutterMethodChannel(name: "kameleoon_client_flutter", binaryMessenger: registrar.messenger())
        let instance = SwiftKameleoonClientFlutterPlugin(channel: channel)
        registrar.addMethodCallDelegate(instance, channel: channel)
    }

    public func handle(_ call: FlutterMethodCall, result: @escaping FlutterResult) {
        let arguments = call.arguments as? NSDictionary
        let siteCode = arguments?["siteCode"] as? String
        let visitorCode = arguments?["visitorCode"] as? String

        if (call.method.elementsEqual("initializeKameleoonClient")) {
            let kameleoonClient = KameleoonClientFactory.create(sitecode : siteCode!)
            clients[siteCode!] = kameleoonClient
            result("")
        } else if (call.method.elementsEqual("triggerExperiment")) {
            let experimentID = arguments?["experimentID"] as? Int
            do {
                try result(clients[siteCode!]?.triggerExperiment(visitorCode : visitorCode!, experimentID : experimentID!))
            } catch {
                result(catchError(error))
            }
        } else if (call.method.elementsEqual("flush")) {
            do {
                if (visitorCode != nil){
                    try clients[siteCode!]?.flush(visitorCode: visitorCode!)
                } else {
                    try clients[siteCode!]?.flush()
                }
                result("")
            } catch {
                result(catchError(error))
            }
        } else if (call.method.elementsEqual("trackConversion")) {
          let goalID = arguments?["goalID"] as? Int
          let revenue = arguments?["revenue"] as? Double
            do {
                try clients[siteCode!]?.trackConversion(visitorCode: visitorCode!, goalId: goalID!, revenue: Float(revenue ?? 0))
                result("")
            } catch {
                result(catchError(error))
            }
        } else if (call.method.elementsEqual("activateFeature")) {
          let featureKey = arguments?["featureKey"] as? String
          let featureID = arguments?["featureID"] as? Int
          do {
            let feature : Bool
            if (featureID != nil) {
                print("Please use `featureKeyOrID` with type of `String`. This is necessary to support multi-environment feature. Supporting of `Int` type will be removed in next releases")
                try feature = clients[siteCode!]?.activateFeature(visitorCode : visitorCode!, featureID : featureID!) == true
            }
            else {
                try feature = clients[siteCode!]?.activateFeature(visitorCode : visitorCode!, featureKey : featureKey!) == true
            }
            result(feature)
          } catch {
            result(catchError(error))
          }
        } else if (call.method.elementsEqual("obtainVisitorCode")) {
          let defaultVisitorCode = arguments?["defaultVisitorCode"] as? String
          if (defaultVisitorCode != nil) {
            result(defaultVisitorCode)
          } else {
            result(UUID().uuidString)
          }
        } else if (call.method.elementsEqual("obtainVariationAssociatedData")) {
          let variationID = arguments?["variationID"] as? Int
          do {
              let data = try clients[siteCode!]?.obtainVariationAssociatedData(variationID : variationID!)
              result(data)
          } catch {
              result(catchError(error))
          }
        } else if (call.method.elementsEqual("obtainFeatureVariable")) {
            let featureKey = arguments?["featureKey"] as? String
            let featureID = arguments?["featureID"] as? Int
            let variableKey = arguments?["variableKey"] as? String
            do {
                if (featureID != nil)
                {
                  print("Please use `featureKeyOrID` with type of `String`. This is necessary to support multi-environment feature. Supporting of `Int` type will be removed in next releases")
                  try result(clients[siteCode!]?.obtainFeatureVariable(featureID : featureID!, variableKey : variableKey!))
                }
                else
                {
                  try result(clients[siteCode!]?.obtainFeatureVariable(featureKey : featureKey!, variableKey : variableKey!))
                }
            }
            catch {
                result(catchError(error))
            }
        } else if (call.method.elementsEqual("addData")) {
            let dataType = arguments?["dataType"] as? String
            
            let data: KameleoonData?
            switch dataType {
            case "CustomData":
                let id = arguments?["id"] as? Int
                let value = arguments?["value"] as? String
                data = CustomData(id: id!, value: value!)
            case "PageView":
                let url = arguments?["url"] as? String
                let title = arguments?["title"] as? String
                let referrer = arguments?["referrer"] as? Int
                data = PageView(url: url!, title: title!, referrer: referrer!)
            case "Interest":
                let index = arguments?["index"] as? Int
                data = Interest(index: index!)
            case "Conversion":
                let goalId = arguments?["goalId"] as? Int
                let revenue = arguments?["revenue"] as? Float
                let negative = arguments?["negative"] as? Bool
                data = Conversion(goalId : goalId!, revenue: revenue!, negative: negative!)
            case "Browser":
                let browser = arguments?["browser"] as? Int
                data = Browser.init(rawValue: browser!)
            default:
                data = nil
                print("Data type not compatible")
            }
            do {
                if (data != nil) {
                    try clients[siteCode!]?.addData(visitorCode: visitorCode!, data: data!);
                }
                result(true)
            } catch {
                result(catchError(error))
            }
        } else if (call.method.elementsEqual("runWhenReady")) {
            let timeout = arguments?["timeout"] as? Int ?? 2_000
            clients[siteCode!]?.runWhenReady(timeoutMilliseconds: timeout) { success in
                if (success) {
                    print("ready")
                    self.channel.invokeMethod("setReady", arguments: true)
                }
                else {
                    print("not ready")
                    self.channel.invokeMethod("setReady", arguments: false)
                }
            }
            result("")
        } else if (call.method.elementsEqual("retrieveDataFromRemoteSource")) {
            let key = arguments?["key"] as? String ?? ""
            do {
                try clients[siteCode!]?.retrieveDataFromRemoteSource(key: key) { [weak self] (data: Data) in
                    do {
                        if let json = try JSONSerialization.jsonObject(with: data) as? [String: Any] {
                            result(json)
                        }
                    } catch {
                        result(self?.catchError(error))
                    }
                }
            } catch {
                result(catchError(error))
            }
        }
    }
    
    private func catchError(_ error: Error) -> FlutterError {
        
        enum VisitorCodeError: String {
            case empty = "visitorCodeEmpty", tooLong = "visitorCodeTooLong"
        }
        
        let errorCode: String
        switch error {
        case KameleoonError.experimentConfigurationNotFound:
            errorCode = "ExperimentConfigurationNotFound"
        case KameleoonError.featureConfigurationNotFound:
            errorCode = "FeatureConfigurationNotFound"
        case KameleoonError.sdkNotReady:
            errorCode = "SDKNotReady"
        case KameleoonError.notTargeted:
            errorCode = "NotTargeted"
        case KameleoonError.notActivated:
            errorCode = "NotActivated"
        case KameleoonError.variationConfigurationNotFound:
            errorCode = "VariationConfigurationNotFound"
        case KameleoonError.visitorCodeNotValid:
            errorCode = "VisitorCodeNotValid"
        case KameleoonError.siteCodeDisabled:
            errorCode = "SiteCodeDisabled"
        default:
            errorCode = "ERROR"
        }
        return FlutterError.init(
            code: errorCode,
            message: error.localizedDescription,
            details: nil)
    }
}
